import importlib
import inspect
import logging
import types

from configparser import ConfigParser
from pathlib import Path

def is_float(s):
    try:
        float(s)
        return True
    except ValueError:
        return False

def is_bool(s):
    return s.lower() in ['true', 'false', 't', 'f']

def destringify_input_params(in_dict):
    """
    Converts input strings from config file to appropriate types.
    """
    for k, v in in_dict.items():
        if is_float(v):
            in_dict[k] = float(v)
        elif is_bool(v):
            in_dict[k] = v.lower() in ['true', 't']  # Convert to boolean directly
    return in_dict

def get_input_output_dirs(config_file: str = "io_dirs.ini", log_level: str = "INFO") -> dict:
    """
    Reads a config file to retrieve input & output dirs & filess.

    Parameters:
    config_file: Path to the configuration file.
    log_level: Logging level as a string (DEBUG, INFO, WARNING, ERROR, CRITICAL).

    Returns:
    dict with input/output paths & files
    """
    io_dir_dict = dict()
    
    # Configure logging level
    log_level = log_level.upper()
    valid_levels = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
    if log_level not in valid_levels:
        raise ValueError(f"Invalid log_level: {log_level}. Valid levels are {valid_levels}.")
    
    logging.basicConfig(level=getattr(logging, log_level))
    logging.info(f"Log level set to: {log_level}")
    
    # Initialize config parser and read the config file
    config = ConfigParser()
    config.read(config_file)

    if not config.sections():
        logging.warning(f"No sections found in the config file {config_file}.")
        return []

    # Iterate over sections (io dirs) in the config file
    for section in config.sections():
        name_str = section
        io_dir_dict[section] = dict(config.items(section))  

    # Create and store Path to src data
    src_dir = Path(io_dir_dict["InputDirs"]["data_root_dir"]) / Path(io_dir_dict["InputDirs"]["dataset"])
    io_dir_dict["InputDirs"]["src_dir"] = src_dir

    # Ensure output dir for frames exists
    out_dir = Path(io_dir_dict["OutputDirs"]["output_root_dir"])
    out_dir.mkdir(exist_ok=True, parents=True)
    io_dir_dict["OutputDirs"]["output_root_dir"] = out_dir
    logging.info(f"Output Dir for frames: {out_dir}")

    # Ensure output dir for movies exists
    movie_output_root_dir = out_dir / Path("Movies") / Path(io_dir_dict["InputDirs"]["dataset"])
    movie_output_root_dir.mkdir(exist_ok=True, parents=True)
    io_dir_dict["OutputDirs"]["movie_output_root_dir"] = movie_output_root_dir 
    logging.info(f"Output Dir for movies: {movie_output_root_dir }")

    return io_dir_dict

    
def get_filter_functions(config_file: str = "filters.ini", log_level: str = "INFO") -> list[tuple[callable, str, dict]]:
    """
    Reads a config file to retrieve image processing filters and their parameters.

    Parameters:
    config_file: Path to the configuration file.
    log_level: Logging level as a string (DEBUG, INFO, WARNING, ERROR, CRITICAL).

    Returns:
    list of tuples: Each tuple is of form (filter_function, name_str, params).
    """
    
    # Configure logging level
    log_level = log_level.upper()
    valid_levels = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
    if log_level not in valid_levels:
        raise ValueError(f"Invalid log_level: {log_level}. Valid levels are {valid_levels}.")
    
    logging.basicConfig(level=getattr(logging, log_level))
    logging.info(f"Log level set to: {log_level}")
    
    # Initialize config parser and read the config file
    config = ConfigParser()
    config.read(config_file)

    if not config.sections():
        logging.warning(f"No sections found in the config file {config_file}.")
        return []

    # List of filter functions and parameters from config file
    filter_function_list = []

    # Cache for imported modules to avoid redundant imports
    module_cache = {}

    # Iterate over sections (filter names) in the config file
    for section in config.sections():
        name_str = section
        params = dict(config.items(section))  # Get parameters for the filter as a dictionary
        filter_name = params.pop('filter_name', name_str)
        params = destringify_input_params(params)

        logging.info(f"Processing filter: {filter_name} with params: {params}")

        # Split the filter_name into module and function parts if it's fully qualified
        module_name, func_name = None, None
        if '.' in filter_name:
            *module_parts, func_name = filter_name.split('.')
            module_name = '.'.join(module_parts)

            # Cache the module import
            if module_name not in module_cache:
                try:
                    module_cache[module_name] = importlib.import_module(module_name)
                except ImportError as e:
                    logging.error(f"Failed to import module {module_name} for filter {name_str}: {e}")
                    raise ImportError(f"Failed to import module: {module_name} for filter {name_str}")

            module = module_cache[module_name]

            try:
                filter_func = getattr(module, func_name)
            except AttributeError as e:
                logging.error(f"Function {func_name} not found in module {module_name} for filter {name_str}: {e}")
                raise AttributeError(f"Function {func_name} not found in module: {module_name} for filter {name_str}")

        else:
            # If no module is provided, the function is assumed to be in the global namespace
            filter_func = globals().get(filter_name)
            if filter_func is None:
                logging.error(f"Unknown filter: {filter_name}")
                raise ValueError(f"Unknown filter: {filter_name}")

        if isinstance(filter_func, type) and callable(filter_func):
            logging.info(f"{filter_func.__name__} is Callable Class")
            logging.info(f"    Init args: {inspect.signature(filter_func.__init__)}")
            init_args = inspect.signature(filter_func.__init__).parameters
            init_params_dict = {}
            for init_param_name, param in init_args.items():
                if init_param_name != 'self':  # Skip 'self'
                    if param.kind in [inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD]:
                        logging.warning(f"Function {filter_name} has variadic args ({param.kind}), which are not handled.")
                        continue
                    if param.default is inspect.Parameter.empty:  # No default, so it's a required parameter
                        # If the parameter is missing, raise an error
                        if init_param_name not in params:
                            raise ValueError(f"Missing required parameter: {init_param_name} to initialize {filter_name}")
                        init_params_dict[init_param_name] = params.pop(init_param_name)
                    else:
                        # Use the parameter from 'params' if available, otherwise use the default
                        init_params_dict[init_param_name] = params.pop(init_param_name, param.default)
            filter_func = filter_func(**init_params_dict)

        elif isinstance(filter_func, types.FunctionType):
            logging.info(f"{filter_func.__name__} is a Function")

        # Add function & params to list
        filter_function_list.append((filter_func, name_str, params))

    return filter_function_list