import cv2
import numpy as np
from pathlib import Path
from predictor_utils import Predictor

from scipy.ndimage import uniform_filter
from PIL import Image, ImageEnhance
from skimage import exposure


def np_stats_str(arr):
    """Returns a string summarizing min, mean, and max values of a NumPy array."""
    return f"{np.min(arr):.4f}, {np.mean(arr):.4f}, {np.max(arr):.4f}"

'''
def get_grid_cell(num_rows, num_cols, grid_size, curr_row, curr_col):

    # Determine start & end rows for current grid cell
    if curr_row + grid_size <= num_rows:
        start_row = curr_row
        end_row = start_row + grid_size
    else:
        start_row = curr_row #num_rows - grid_size
        end_row = num_rows

    # Determine start and end cols for current grid cell
    if curr_col + grid_size <= num_cols:
        start_col = curr_col
        end_col = start_col + grid_size
    else:
        start_col = curr_col #num_cols - grid_size
        end_col = num_cols

    return start_col, end_col, start_row, end_row
'''

def get_grid_cell(num_rows, num_cols, grid_size, curr_row, curr_col):
    """Determine the grid cell's start and end positions based on current row and column."""
    start_row = curr_row
    end_row = min(start_row + grid_size, num_rows)
    start_col = curr_col
    end_col = min(start_col + grid_size, num_cols)
    return start_col, end_col, start_row, end_row


def identity_filter(image_np):
    """Return the image without applying any modifications."""
    return image_np



class Delete_Big_Fish:
    """Class to detect and remove large fish from an image using bounding boxes."""


    def __init__(self, expt_cfg, model_weights, conf_thresh, grid_size, pad=10):
        self.predictor = Predictor(expt_cfg, model_weights, conf_thresh)
        self.grid_size = 50
        self.ctr = 0
        self.pad = 10

    def __call__(self, image_np):
        """Main method to process an image and remove large fish based on bounding box predictions."""
        pred_boxes = self.predictor.get_pred_bboxes(image_np)
        #out_dir =  Path("./new_filter_movies") / Path("pred_boxes")
        #out_file = out_dir / Path("Frame_"+str(self.ctr)+".npz")
        #out_dir.mkdir(exist_ok=True, parents=True)
        #np.savez(out_file, *pred_boxes)
        #self.ctr += 1

        big_fish_free_np = self.delete_big_fish(image_np, pred_boxes)
        return big_fish_free_np

    
    def delete_big_fish(self, image_np, pred_boxes):
        """
        Remove large fish from the image by replacing the pixel values in detected bounding boxes
        with the average pixel values from the surrounding grid or bounding box region.
    
        Parameters:
        image_np (np.ndarray): Input image with shape (height, width, 3), where channels represent (Raw, Background, Frame Diff).
        pred_boxes (list): List of bounding box coordinates in the format [(x1, y1, x2, y2), ...].
        grid_size (int): Size of the grid over which to compute the average pixel values. If grid_size=0, averaging is done over the bounding box.
    
        Returns:
        np.ndarray: Image with large fish removed (replaced with average pixel values).
        """
    
        # Copy the image to avoid modifying the original
        big_fish_free = np.copy(image_np)
        
        # Separate the color channels (Raw, Background, Frame Diff)
        raw_channel = big_fish_free[:, :, 0]
        bckgnd_channel = big_fish_free[:, :, 1]
        dframe_channel = big_fish_free[:, :, 2]
    
        # Iterate through the list of detected bounding boxes
        for curr_pred_box in pred_boxes:
            x1, y1, x2, y2 = curr_pred_box
                        
            if self.grid_size > 0:
                # Calculate grid-aligned coordinates
                x1b = int(self.grid_size * np.floor(x1 / self.grid_size))
                x2b = int(self.grid_size * np.ceil(x2 / self.grid_size))
                y1b = int(self.grid_size * np.floor(y1 / self.grid_size))
                y2b = int(self.grid_size * np.ceil(y2 / self.grid_size))
    
                # Compute mean pixel values over the grid region for each channel
                region = big_fish_free[y1b:y2b, x1b:x2b, :]
                raw_chan_mean = np.mean(raw_channel[y1b:y2b,x1b:x2b])
                bckgnd_chan_mean = np.mean(bckgnd_channel[y1b:y2b,x1b:x2b])
                dframe_chan_mean = np.mean(dframe_channel[y1b:y2b,x1b:x2b])
    
            else:
                # Compute mean pixel values directly over the bounding box
                region = big_fish_free[y1:y2, x1:x2, :]
                raw_chan_mean = np.mean(raw_channel[y1b:y2b,x1b:x2b])
                bckgnd_chan_mean = np.mean(bckgnd_channel[y1b:y2b,x1b:x2b])
                dframe_chan_mean = np.mean(dframe_channel[y1b:y2b,x1b:x2b])

            if self.pad != 0:
                row_max, col_max, _ = big_fish_free.shape
                y1 = max(0, y1-self.pad)
                x1 = max(0, x1-self.pad)
                y2 = min(row_max, y2+self.pad)
                x2 = min(col_max, x2+self.pad)
    
            # Replace detected fish with average pixel values
            raw_channel [int(np.floor(y1)):int(np.ceil(y2)),
                         int(np.floor(x1)):int(np.ceil(x2))] = raw_chan_mean
            bckgnd_channel [int(np.floor(y1)):int(np.ceil(y2)),
                            int(np.floor(x1)):int(np.ceil(x2))] = bckgnd_chan_mean
            dframe_channel [int(np.floor(y1)):int(np.ceil(y2)),
                            int(np.floor(x1)):int(np.ceil(x2))] = dframe_chan_mean
              
        big_fish_free_np = np.stack((raw_channel, bckgnd_channel, dframe_channel), axis=-1)
    
        return big_fish_free_np

def select_color_frame(image_np, frame):
    frame=int(frame)
    return image_np[:,:,frame]

def remove_dark_noise(image_np, threshold_perc):

    threshold_perc = int(threshold_perc)
    thresh = np.percentile(image_np[:,:], threshold_perc)
    image_np[:,:][image_np[:,:] < thresh] = int(thresh)

    return image_np


def guided_filter_channel(I, p, radius, eps):
    """
    Guided filter applied to a single channel.
    Good for smoothing images and preserving edges
    
    Parameters:
    - I: Guidance image (H, W), single channel.
    - p: Input image to be filtered (H, W), single channel.
    - radius: Radius of the filter window.
    - eps: Regularization term.
    
    Returns:
    - q: Filtered output image (H, W).
    """
    mean_I = uniform_filter(I, size=radius)
    mean_p = uniform_filter(p, size=radius)
    mean_Ip = uniform_filter(I * p, size=radius)
    cov_Ip = mean_Ip - mean_I * mean_p

    mean_II = uniform_filter(I * I, size=radius)
    var_I = mean_II - mean_I * mean_I

    a = cov_Ip / (var_I + eps)
    b = mean_p - a * mean_I

    mean_a = uniform_filter(a, size=radius)
    mean_b = uniform_filter(b, size=radius)

    q = mean_a * I + mean_b
    return q

def self_guided_filter(image_np, radius=5, eps=0.01):
    """
    Apply vectorized guided filtering independently to each channel of an RGB image.

    Parameters:
    - img: Input image (H, W)
    - radius: Radius of the filter
    - eps: Regularization term
    
    Returns:
    - Filtered image (H, W)
    """
    image_np = image_np.astype(np.float32)  # Convert image to float32 for computation
    
    # Initialize output image
    filtered_image_np = np.zeros_like(image_np)
    filtered_image_np = guided_filter_channel(image_np, image_np, radius, eps)
    filtered_image_np = np.clip(filtered_image_np, 0, 255).astype(np.uint8)

    return filtered_image_np


def image_enhance_1D(image_np, enhance_factor):

    # Convert numpy to PIL Images
    img_pil = Image.fromarray(image_np.astype(np.uint8))
    
    # Apply ImageEnhance to each CLAHE-enhanced channel
    enhancer_pil = ImageEnhance.Contrast(img_pil)
    enhanced_pil = enhancer_pil.enhance(enhance_factor)
    
    # Convert enhanced PIL Images back to NumPy arrays
    enhanced_pil_np = np.array(enhanced_pil)

    return enhanced_pil

def nth_largest_element_wrapper(n):
    """
    Create a function that returns the nth largest element from a NumPy array.
    
    Parameters:
    - n: The position of the largest value to find (e.g., 1 for largest, 2 for second largest).
    
    Returns:
    - A function that takes a NumPy array as input and returns the nth largest element.
    """
    def nth_largest(arr):
        """
        Find the nth largest value in the array using np.partition.
        
        Parameters:
        - arr: Input NumPy array (1D).
        
        Returns:
        - The nth largest value.
        """
        if not (1 <= n <= len(arr)):
            raise ValueError("n must be between 1 and the length of the array.")
        
        # Perform partitioning to find the nth largest element
        partitioned_arr = np.partition(arr.ravel(), -n)
        
        # Return the nth largest value
        return partitioned_arr[-n]
    
    return nth_largest


def np_max_filter_function(image_np, cardinal_thresh, grid_size, threshold, 
                           enhance_factor, radius, eps):
    """
    Applies a grid-based max filter to an image, enhancing or smoothing grid cells based on a threshold.
    
    Parameters:
    image_np (np.ndarray): Input image as a NumPy array.
    cardinal_thresh (int): The 'n' value for finding the nth largest element in each grid cell.
    grid_size (int): Size of each grid cell.
    threshold (float): The value used to decide if a grid cell should be enhanced or smoothed.
    enhance_factor (float): Factor by which to enhance the contrast of a grid cell.
    radius (int): Radius for the guided filter.
    eps (float): Epsilon parameter for the guided filter.

    Returns:
    np.ndarray: The processed image after filtering.
    """

    # Convert image to float64 for accurate processing
    image_np = image_np.astype(np.float64)
    
    # Create a copy of the input image to store the processed result
    new_image_np = np.copy(image_np)
    
    # Wrapper to find the nth largest element in an array (used for thresholding)
    np_nth_max = nth_largest_element_wrapper(int(cardinal_thresh))
    
    # Convert grid_size to integer if not already
    grid_size = int(grid_size)

    # Get the dimensions of the image
    num_rows, num_cols = image_np.shape
    
    # Loop through the image in grid-sized steps
    for curr_row in range(0, num_rows, grid_size):
        for curr_col in range(0, num_cols, grid_size):
    
            # Determine the start and end row for the current grid cell
            # Ensure that the grid cell does not exceed image bounds
            if curr_row + grid_size <= num_rows:
                start_row = curr_row
                end_row = start_row + grid_size
            else:
                start_row = curr_row
                end_row = num_rows  # Limit to image boundary
        
            # Determine the start and end column for the current grid cell
            if curr_col + grid_size <= num_cols:
                start_col = curr_col
                end_col = start_col + grid_size
            else:
                start_col = curr_col
                end_col = num_cols  # Limit to image boundary

            # Extract the current grid cell from the image
            curr_cell = image_np[start_row:end_row, start_col:end_col]
    
            # Get the nth largest value in the current grid cell
            curr_filter_param = np_nth_max(curr_cell[:,:])
    
            # If the nth largest value exceeds the threshold, enhance the contrast of the cell
            if curr_filter_param > threshold:
                new_image_np[start_row:end_row, start_col:end_col] = image_enhance_1D(
                    curr_cell[:,:], enhance_factor=enhance_factor)
            else:
                # Otherwise, smooth the cell using the guided filter
                # First, calculate the mean of the grid cell
                new_image_np[start_row:end_row, start_col:end_col] = np.mean(curr_cell[:,:])
                
                # Apply the guided filter for smoothing
                new_image_np[start_row:end_row, start_col:end_col] = \
                    self_guided_filter(curr_cell[:,:], radius=radius, eps=eps)

    # Convert the result back to uint8 to ensure valid image data type
    new_image_np = np.clip(new_image_np, 0, 255).astype(np.uint8)
    
    return new_image_np


def clahe_filter(image_np, clip_limit):

    # Apply CLAHE to each channel
    clahe_pil = exposure.equalize_adapthist(image_np, clip_limit=clip_limit)

    # Convert back to 3Plane Numpy
    clahe_np = np.array(clahe_pil)
    clahe_np *= 255
    clahe_np = np.clip(clahe_np, 0, 255).astype(np.uint8)

    return clahe_np


def binarize_image(image, threshold):
    
    # Set values below switch_level to 0 and others to 255
    image[image < threshold] = 0
    image[image >= threshold] = 255
    image = np.clip(image, 0, 255).astype(np.uint8)

    return image


def fill_in_contours(image):

    # Find contours of the inverted image (external contours)
    contours, _ = cv2.findContours(image, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

    # Draw contours on the binary image to close holes
    image_copy = image.copy()
    cv2.drawContours(image_copy, contours, -1, 255, thickness=cv2.FILLED)

    return image_copy
    

def get_gradient_magnitude_image(image):
    
    # Apply Sobel operators
    sobel_x = cv2.Sobel(image, cv2.CV_64F, 1, 0, ksize=3)
    sobel_y = cv2.Sobel(image, cv2.CV_64F, 0, 1, ksize=3)
    
    # Compute and normalize the gradient magnitude
    gradient_magnitude = np.sqrt(sobel_x**2 + sobel_y**2)
    gradient_magnitude = cv2.normalize(gradient_magnitude, None, 0, 255, cv2.NORM_MINMAX)

    return gradient_magnitude.astype(np.uint8)

def median_filter_torch(image, kernel_size=3):
    """ Median filter implementation using PyTorch conv2d for extracting patches """
    
    # Ensure image is a PyTorch tensor
    if not isinstance(image, torch.Tensor):
        image = torch.tensor(image)
    
    # Ensure image is float type and add batch and channel dimensions if necessary
    image = image.float()
    if len(image.shape) == 2:  # grayscale image
        image = image.unsqueeze(0).unsqueeze(0)  # Add batch and channel dimensions
    elif len(image.shape) == 3:  # RGB or multi-channel image
        image = image.unsqueeze(0)  # Add batch dimension
    
    
    # Create a filter that extracts patches using conv2d with identity kernel
    channels = image.shape[1]
    kernel = torch.eye(channels).view(channels, channels, 1, 1)

    #  Move to the GPU (if available)
    if torch.cuda.is_available():
        image = image.cuda()
        kernel=kernel.cuda()


    # Apply padding to handle the border pixels
    kernel_size = int(kernel_size)
    padding = kernel_size // 2
    #print(f"median image dim: {image.ndim}, { (padding, padding, padding, padding)}")
    image_padded = F.pad(image, (padding, padding, padding, padding), mode='reflect')

    # Use conv2d to extract sliding windows (patches)
    patches = F.conv2d(image_padded, kernel, stride=1, padding=0, groups=channels)
    patches = patches.unfold(2, kernel_size, 1).unfold(3, kernel_size, 1)

    # Compute the median across the kernel dimension
    median_filtered = torch.median(patches.contiguous().view(*patches.shape[:4], -1), dim=-1)[0]

    # Return to the original shape and move back to CPU
    output = median_filtered.squeeze().cpu().numpy().astype(np.uint8)

    return output
