import cv2
import numpy as np
from pathlib import Path
from predictor_utils import Predictor

from scipy.ndimage import uniform_filter


def np_stats_str(arr):
    """Returns a string summarizing min, mean, and max values of a NumPy array."""
    return f"{np.min(arr):.4f}, {np.mean(arr):.4f}, {np.max(arr):.4f}"

def get_grid_cell(num_rows, num_cols, grid_size, curr_row, curr_col):
    """Determine the grid cell's start and end positions based on current row and column."""
    start_row = curr_row
    end_row = min(start_row + grid_size, num_rows)
    start_col = curr_col
    end_col = min(start_col + grid_size, num_cols)
    return start_col, end_col, start_row, end_row

def identity_filter(image_np):
    """Return the image without applying any modifications."""
    return image_np

class Delete_Big_Fish:
    """Class to detect and remove large fish from an image using bounding boxes."""

    def __init__(self, expt_cfg, model_weights, conf_thresh, grid_size):
        self.predictor = Predictor(expt_cfg, model_weights, conf_thresh)
        self.grid_size = grid_size
        self.ctr = 0

    def __call__(self, image_np):
        """Main method to process an image and remove large fish based on bounding box predictions."""
        pred_boxes = self.predictor.get_gnd_truth_bboxes(image_np)
        out_dir = Path("./new_filter_movies/pred_boxes")
        out_dir.mkdir(exist_ok=True, parents=True)
        np.savez(out_dir / f"Frame_{self.ctr}.npz", *pred_boxes)
        self.ctr += 1
        return self.delete_big_fish(image_np, pred_boxes)

    def delete_big_fish(self, image_np, pred_boxes):
        """
        Replace pixel values within the bounding boxes of detected fish with average pixel values.
        This removes large fish from the image.
        """
        big_fish_free = np.copy(image_np)  # Copy to avoid modifying the original image
        
        # Separate the channels (Raw, Background, Frame Diff)
        raw_channel = big_fish_free[:, :, 0]
        bckgnd_channel = big_fish_free[:, :, 1]
        dframe_channel = big_fish_free[:, :, 2]

        # Process each bounding box
        for x1, y1, x2, y2 in pred_boxes:
            if self.grid_size > 0:
                # Align bounding box with grid size
                x1b = int(self.grid_size * np.floor(x1 / self.grid_size)) 
                x2b = int(self.grid_size * np.ceil(x2 / self.grid_size))
                y1b = int(self.grid_size * np.floor(y1 / self.grid_size)) 
                y2b = int(self.grid_size * np.ceil(y2 / self.grid_size))
                
                raw_chan_mean = np.mean(raw_channel[y1b:y2b, x1b:x2b])
                bckgnd_chan_mean = np.mean(bckgnd_channel[y1b:y2b, x1b:x2b])
                dframe_chan_mean = np.mean(dframe_channel[y1b:y2b, x1b:x2b])
            else:
                # Use the bounding box itself if grid size is 0
                raw_chan_mean = np.mean(raw_channel[y1:y2, x1:x2])
                bckgnd_chan_mean = np.mean(bckgnd_channel[y1:y2, x1:x2])
                dframe_chan_mean = np.mean(dframe_channel[y1:y2, x1:x2])

            # Replace detected fish region with average pixel values
            raw_channel [int(np.floor(y1)):int(np.ceil(y2)),
                         int(np.floor(x1)):int(np.ceil(x2))] = raw_chan_mean
            bckgnd_channel [int(np.floor(y1)):int(np.ceil(y2)),
                            int(np.floor(x1)):int(np.ceil(x2))] = bckgnd_chan_mean
            dframe_channel [int(np.floor(y1)):int(np.ceil(y2)),
                            int(np.floor(x1)):int(np.ceil(x2))] = dframe_chan_mean

        return np.stack((raw_channel, bckgnd_channel, dframe_channel), axis=-1)

def select_color_frame(image_np, frame):
    """Select a specific frame from a multi-frame image."""
    return image_np[:, :, int(frame)]

def remove_dark_noise(image_np, threshold_perc):
    """Remove dark noise by thresholding pixel values below a certain percentile."""
    thresh = np.percentile(image_np, threshold_perc)
    image_np[image_np < thresh] = int(thresh)
    return image_np

def guided_filter_channel(I, p, radius, eps):
    """
    Apply a guided filter to a single image channel.
    I: Guidance image, p: Input image to filter, radius: Filter window radius, eps: Regularization.
    """
    mean_I = uniform_filter(I, size=radius)
    mean_p = uniform_filter(p, size=radius)
    mean_Ip = uniform_filter(I * p, size=radius)
    cov_Ip = mean_Ip - mean_I * mean_p
    mean_II = uniform_filter(I * I, size=radius)
    var_I = mean_II - mean_I * mean_I
    a = cov_Ip / (var_I + eps)
    b = mean_p - a * mean_I
    mean_a = uniform_filter(a, size=radius)
    mean_b = uniform_filter(b, size=radius)
    return mean_a * I + mean_b

def self_guided_filter(image_np, radius=5, eps=0.01):
    """Apply a guided filter independently to each channel of an image."""
    return np.clip(guided_filter_channel(image_np, image_np, radius, eps), 0, 255).astype(np.uint8)

def image_enhance_1D(image_np, enhance_factor):
    """Enhance contrast of a single channel image using PIL."""
    img_pil = Image.fromarray(image_np.astype(np.uint8))
    enhancer = ImageEnhance.Contrast(img_pil)
    enhanced_pil = enhancer.enhance(enhance_factor)
    return np.array(enhanced_pil)

def nth_largest_element_wrapper(n):
    """Return a function to find the nth largest element in a NumPy array."""
    def nth_largest(arr):
        return np.partition(arr.ravel(), -n)[-n]
    return nth_largest

def np_max_filter_function(image_np, cardinal_thresh, grid_size, threshold, 
                           enhance_factor, radius, eps):
    """
    Apply a maximum filter and guided filtering to enhance grid cells in an image.
    """
    np_nth_max = nth_largest_element_wrapper(int(cardinal_thresh))
    new_image_np = np.copy(image_np)

    num_rows, num_cols = image_np.shape
    for curr_row in range(0, num_rows, grid_size):
        for curr_col in range(0, num_cols, grid_size):
            start_col, end_col, start_row, end_row = get_grid_cell(num_rows, num_cols, grid_size, curr_row, curr_col)
            curr_cell = image_np[start_row:end_row, start_col:end_col]
            curr_filter_param = np_nth_max(curr_cell)
            if curr_filter_param > threshold:
                new_image_np[start_row:end_row, start_col:end_col] = image_enhance_1D(curr_cell, enhance_factor)
            else:
                new_image_np[start_row:end_row, start_col:end_col] = self_guided_filter(curr_cell, radius, eps)

    return new_image_np.astype(np.uint8)

def clahe_filter(image_np, clip_limit):
    """Apply CLAHE (Contrast Limited Adaptive Histogram Equalization) to enhance contrast."""
    clahe_np = exposure.equalize_adapthist(image_np, clip_limit=clip_limit)
    return (clahe_np * 255).astype(np.uint8)

def binarize_image(image, threshold):
    """Binarize an image based on a threshold, setting values below the threshold to 0 and above to 255."""
    image[image < threshold] = 0
    image[image >= threshold] = 255
    return np.clip(image, 0, 255).astype(np.uint8)

def fill_in_contours(image):
    """Fill in contours of a binary image to close gaps."""
    contours, _ = cv2.findContours(image, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    image_copy = image.copy()
    cv2.drawContours(image_copy, contours, -1, 255, thickness=cv2.FILLED)
    return image_copy

def get_gradient_magnitude_image(image):
    """Compute the gradient magnitude of an image using Sobel operators."""
    sobel_x = cv2.Sobel(image, cv2.CV_64F, 1, 0, ksize=3)
    sobel_y = cv2.Sobel(image, cv2.CV_64F, 0, 1, ksize=3)
    gradient_magnitude = np.sqrt(sobel_x**2 + sobel_y**2)
    return cv2.normalize(gradient_magnitude, None, 0, 255, cv2.NORM_MINMAX).astype(np.uint8)


