import cv2
import inspect
import logging
import matplotlib.pyplot as plt
import numpy as np
import sys
import torch
import torch.nn.functional as F

from collections import defaultdict
from IPython.display import display
from pathlib import Path
from PIL import Image, ImageDraw, ImageEnhance
from scipy.ndimage import gaussian_filter, uniform_filter
from skimage import exposure, io, restoration
from skimage.filters import median
from skimage.morphology import disk
from skimage.util import view_as_blocks, view_as_windows
from time import time

# import some common detectron2 utilities
from detectron2.config import get_cfg
from detectron2.engine import DefaultPredictor

# Get ALDI imports
from aldi.config import add_aldi_config

from predictor_utils import Predictor

# Big Fish predictor is global variable. This way the
# can retain deep net used for locating big fish 
# after it has been loaded and not need to reload
# in subsequent calls to method for big fish detection
bf_predictor = None
logging.basicConfig(level=logging.INFO)

def np_stats_str(arr):
    return f"{np.min(arr):.4f}, {np.mean(arr):.4f}, {np.max(arr):.4f}"



def get_grid_cell(num_rows, num_cols, grid_size, curr_row, curr_col):

    # Determine start & end rows for current grid cell
    if curr_row + grid_size <= num_rows:
        start_row = curr_row
        end_row = start_row + grid_size
    else:
        start_row = curr_row #num_rows - grid_size
        end_row = num_rows

    # Determine start and end cols for current grid cell
    if curr_col + grid_size <= num_cols:
        start_col = curr_col
        end_col = start_col + grid_size
    else:
        start_col = curr_col #num_cols - grid_size
        end_col = num_cols

    return start_col, end_col, start_row, end_row


def identity_filter(image_np):
    return image_np



class Delete_Big_Fish:

    def __init__(self, expt_cfg, model_weights, conf_thresh, grid_size):
        self.predictor = Predictor(expt_cfg, model_weights, conf_thresh)
        self.grid_size = 50
        self.ctr = 0

    def __call__(self,image_np):
        pred_boxes = self.predictor.get_gnd_truth_bboxes(image_np)
        out_dir =  Path("./new_filter_movies") / Path("pred_boxes")
        out_file = out_dir / Path("Frame_"+str(self.ctr)+".npz")
        out_dir.mkdir(exist_ok=True, parents=True)
        np.savez(out_file, *pred_boxes)
        self.ctr += 1
        #print(pred_boxes)
        big_fish_free_np = self.delete_big_fish(image_np, pred_boxes)
        return big_fish_free_np

    
    def delete_big_fish(self, image_np, pred_boxes):
        """
        Remove large fish from the image by replacing the pixel values in detected bounding boxes
        with the average pixel values from the surrounding grid or bounding box region.
    
        Parameters:
        image_np (np.ndarray): Input image with shape (height, width, 3), where channels represent (Raw, Background, Frame Diff).
        pred_boxes (list): List of bounding box coordinates in the format [(x1, y1, x2, y2), ...].
        grid_size (int): Size of the grid over which to compute the average pixel values. If grid_size=0, averaging is done over the bounding box.
    
        Returns:
        np.ndarray: Image with large fish removed (replaced with average pixel values).
        """
    
        # Copy the image to avoid modifying the original
        big_fish_free = np.copy(image_np)
        
        # Separate the color channels (Raw, Background, Frame Diff)
        raw_channel = big_fish_free[:, :, 0]
        bckgnd_channel = big_fish_free[:, :, 1]
        dframe_channel = big_fish_free[:, :, 2]
    
        # Iterate through the list of detected bounding boxes
        for curr_pred_box in pred_boxes:
            x1, y1, x2, y2 = curr_pred_box
            
            # Ensure integer coordinates
            #x1, y1, x2, y2 = map(int, [x1, y1, x2, y2])
            
            if self.grid_size > 0:
                # Calculate grid-aligned coordinates
                x1b = int(self.grid_size * np.floor(x1 / self.grid_size))
                x2b = int(self.grid_size * np.ceil(x2 / self.grid_size))
                y1b = int(self.grid_size * np.floor(y1 / self.grid_size))
                y2b = int(self.grid_size * np.ceil(y2 / self.grid_size))
    
                # Compute mean pixel values over the grid region for each channel
                region = big_fish_free[y1b:y2b, x1b:x2b, :]
                raw_chan_mean = np.mean(raw_channel[y1b:y2b,x1b:x2b])
                bckgnd_chan_mean = np.mean(bckgnd_channel[y1b:y2b,x1b:x2b])
                dframe_chan_mean = np.mean(dframe_channel[y1b:y2b,x1b:x2b])
    
            else:
                # Compute mean pixel values directly over the bounding box
                region = big_fish_free[y1:y2, x1:x2, :]
                raw_chan_mean = np.mean(raw_channel[y1b:y2b,x1b:x2b])
                bckgnd_chan_mean = np.mean(bckgnd_channel[y1b:y2b,x1b:x2b])
                dframe_chan_mean = np.mean(dframe_channel[y1b:y2b,x1b:x2b])
    
            # Replace detected fish with average pixel values
            raw_channel [int(np.floor(y1)):int(np.ceil(y2)),
                         int(np.floor(x1)):int(np.ceil(x2))] = raw_chan_mean
            bckgnd_channel [int(np.floor(y1)):int(np.ceil(y2)),
                            int(np.floor(x1)):int(np.ceil(x2))] = bckgnd_chan_mean
            dframe_channel [int(np.floor(y1)):int(np.ceil(y2)),
                            int(np.floor(x1)):int(np.ceil(x2))] = dframe_chan_mean
              
        big_fish_free_np = np.stack((raw_channel, bckgnd_channel, dframe_channel), axis=-1)
        #print(f" Out Delete Big Fish: {np.min(big_fish_free_np[:,:,1])}, {np.max(big_fish_free_np[:,:,1])}")
    
        return big_fish_free_np

def select_color_frame(image_np, frame):
    frame=int(frame)
    return image_np[:,:,frame]

def remove_dark_noise(image_np, threshold_perc):

    threshold_perc = int(threshold_perc)
    thresh = np.percentile(image_np[:,:], threshold_perc)
    image_np[:,:][image_np[:,:] < thresh] = int(thresh)
    #print(int(thresh), threshold_perc, np.min(image_np[:,:]),np.max(image_np[:,:]), image_np[:,:].dtype)

    return image_np


def guided_filter_channel(I, p, radius, eps):
    """
    Guided filter applied to a single channel.
    
    Parameters:
    - I: Guidance image (H, W), single channel.
    - p: Input image to be filtered (H, W), single channel.
    - radius: Radius of the filter window.
    - eps: Regularization term.
    
    Returns:
    - q: Filtered output image (H, W).
    """
    mean_I = uniform_filter(I, size=radius)
    mean_p = uniform_filter(p, size=radius)
    mean_Ip = uniform_filter(I * p, size=radius)
    cov_Ip = mean_Ip - mean_I * mean_p

    mean_II = uniform_filter(I * I, size=radius)
    var_I = mean_II - mean_I * mean_I

    a = cov_Ip / (var_I + eps)
    b = mean_p - a * mean_I

    mean_a = uniform_filter(a, size=radius)
    mean_b = uniform_filter(b, size=radius)

    q = mean_a * I + mean_b
    return q

def self_guided_filter(image_np, radius=5, eps=0.01):
    """
    Apply vectorized guided filtering independently to each channel of an RGB image.

    Parameters:
    - img: Input image (H, W)
    - radius: Radius of the filter
    - eps: Regularization term
    
    Returns:
    - Filtered image (H, W)
    """
    image_np = image_np.astype(np.float32)  # Convert image to float32 for computation
    
    # Initialize output image
    filtered_image_np = np.zeros_like(image_np)
    filtered_image_np = guided_filter_channel(image_np, image_np, radius, eps)
    filtered_image_np = np.clip(filtered_image_np, 0, 255).astype(np.uint8)

    return filtered_image_np





def image_enhance_1D(image_np, enhance_factor):

    # Convert numpy to PIL Images
    img_pil = Image.fromarray(image_np.astype(np.uint8))
    
    # Apply ImageEnhance to each CLAHE-enhanced channel
    enhancer_pil = ImageEnhance.Contrast(img_pil)
    
    enhanced_pil = enhancer_pil.enhance(enhance_factor)
    
    # Convert enhanced PIL Images back to NumPy arrays
    enhanced_pil_np = np.array(enhanced_pil)

    return enhanced_pil

def nth_largest_element_wrapper(n):
    """
    Create a function that returns the nth largest element from a NumPy array.
    
    Parameters:
    - n: The position of the largest value to find (e.g., 1 for largest, 2 for second largest).
    
    Returns:
    - A function that takes a NumPy array as input and returns the nth largest element.
    """
    def nth_largest(arr):
        """
        Find the nth largest value in the array using np.partition.
        
        Parameters:
        - arr: Input NumPy array (1D).
        
        Returns:
        - The nth largest value.
        """
        if not (1 <= n <= len(arr)):
            raise ValueError("n must be between 1 and the length of the array.")
        
        # Perform partitioning to find the nth largest element
        partitioned_arr = np.partition(arr.ravel(), -n)
        
        # Return the nth largest value
        return partitioned_arr[-n]
    
    return nth_largest


def np_max_filter_function(image_np, cardinal_thresh, grid_size, threshold, 
                           enhance_factor, radius, eps):
    
    image_np = image_np.astype(np.float64)
    new_image_np = np.copy(image_np)
    np_nth_max = nth_largest_element_wrapper(int(cardinal_thresh))
    grid_size = int(grid_size)


    # Prepare lists for raw, bckgnd and dframe filter paramss values
    filter_param_lists = []

    # 1st scan to get filter params values
    num_rows, num_cols = image_np.shape
    for curr_row in range(0,num_rows,grid_size):
        for curr_col in range(0,num_cols,grid_size):
    

            # Determine start & end rows for current grid cell
            if curr_row + grid_size <= num_rows:
                start_row = curr_row
                end_row = start_row + grid_size
            else:
                start_row = curr_row #num_rows - grid_size
                end_row = num_rows
        
            # Determine start and end cols for current grid cell
            if curr_col + grid_size <= num_cols:
                start_col = curr_col
                end_col = start_col + grid_size
            else:
                start_col = curr_col #num_cols - grid_size
                end_col = num_cols

    
            curr_cell = image_np[start_row:end_row, start_col:end_col]
    
            #Get filter param for current grid cell
            curr_filter_param = np_nth_max(curr_cell[:,:])
    
            if curr_filter_param > threshold:
                # Increase contrast of entire cell if above threshold
                new_image_np[start_row:end_row, start_col:end_col] = image_enhance_1D(curr_cell[:,:], 
                                                                                         enhance_factor=enhance_factor)
            else:
                new_image_np[start_row:end_row, start_col:end_col] = np.mean(curr_cell[:,:])
                # Smooth cell if beneath threshold
                new_image_np[start_row:end_row, start_col:end_col] = \
                    self_guided_filter(curr_cell[:,:], radius=radius, eps=eps)


    new_image_np = new_image_np.astype(np.uint8)
    return new_image_np           


def clahe_filter(image_np, clip_limit):

    # Apply CLAHE to each channel
    clahe_pil = exposure.equalize_adapthist(image_np, clip_limit=clip_limit)

    # Convert back to 3Plane Numpy
    clahe_np = np.array(clahe_pil)
    clahe_np *= 255
    clahe_np = clahe_np.astype(np.uint8)

    return clahe_np


def binarize_image(image, threshold):
    
    # Set values below switch_level to 0 and others to 255
    image[image < threshold] = 0
    image[image >= threshold] = 255
    image = np.clip(image, 0, 255).astype(np.uint8)

    '''
    # Invert the binary image
    inverted_image = cv2.bitwise_not(image)

    # Find contours of the inverted image (external contours)
    contours, _ = cv2.findContours(inverted_image, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

    # Fill the external contours in the original binary image to close holes
    cv2.drawContours(image, contours, -1, 255, thickness=cv2.FILLED)
    '''
    return image


def fill_in_contours(image):

    # Find contours of the inverted image (external contours)
    contours, _ = cv2.findContours(image, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

    # Draw contours on the binary image to close holes
    #print(" ---- ",image.shape, image.dtype, contours)

    image_copy = image.copy()
    cv2.drawContours(image_copy, contours, -1, 255, thickness=cv2.FILLED)
    #cv2.drawContours(image, contours, -1, 255, thickness=cv2.FILLED)

    # Merge the filled contours with the original binary image using bitwise OR
    #filled_image = cv2.bitwise_or(image, contour_filled_image)
    
    # Fill the external contours in the original binary image to close holes
    #cv2.drawContours(image, contours, -1, 255, thickness=cv2.FILLED)

    return image_copy
    

def get_gradient_magnitude_image(image):
    
    # Apply Sobel operators
    sobel_x = cv2.Sobel(image, cv2.CV_64F, 1, 0, ksize=3)
    sobel_y = cv2.Sobel(image, cv2.CV_64F, 0, 1, ksize=3)
    
    # Compute and normalize the gradient magnitude
    gradient_magnitude = np.sqrt(sobel_x**2 + sobel_y**2)
    gradient_magnitude = cv2.normalize(gradient_magnitude, None, 0, 255, cv2.NORM_MINMAX)

    return gradient_magnitude.astype(np.uint8)


