import argparse
import cv2
import logging
import numpy as np
import operator
import os

from matplotlib import pyplot as plt
from moviepy.editor import ColorClip, clips_array, ImageSequenceClip
from pathlib import Path
from tqdm import tqdm

from config_utils import get_filter_functions, get_input_output_dirs, get_analysis_params

from movie_utils import contour_filter_area, draw_bounding_boxes
from movie_utils import get_contours
from movie_utils import get_padded_bounding_boxes, make_frame_dict, process_single_frame, rel_annotate_frame

from norfair import Detection, Tracker, Video, draw_points, print_objects_as_table


def centered_rolling_average(data, window_size, n):
    if window_size < 2 * n + 1:
        raise ValueError("Window size must be at least 2 * n + 1 to drop n smallest and n largest values.")
    
    half_window = window_size // 2
    result = []
    
    for i in range(len(data)):
        # Define the window bounds
        start = max(0, i - half_window)
        end = min(len(data), i + half_window + 1)
        
        window = data[start:end]
        window_sorted = sorted(window)
        
        # Drop n smallest and n largest if possible
        if len(window_sorted) > 2 * n:
            window_without_extremes = window_sorted[n:-n]
        else:
            window_without_extremes = window_sorted
        
        # Calculate the average and store it
        average = int(np.mean(window_without_extremes))
        result.append(average)
    
    return result

def compare_histograms(list1, list2, title, xlabel, num_bins=50,
                      label1='Fry Present', label2='Fry Absent'):
    """
    Compare two histograms by overlaying them, ensuring both histograms have the same bin edges.
    
    Parameters:
    list1, list2 (list): Lists containing the data to be compared.
    title (str): Title of the plot.
    xlabel (str): Label for the x-axis.
    bins (int or array): Number of bins or bin edges for the histograms.
    """
    
    # Compute the minimum and maximum values across both datasets
    min_value = min(min(list1), min(list2))
    max_value = max(max(list1), max(list2))
    
    # Create a set of bin edges that cover the range of both datasets
    # +1 because linspace includes the end point
    bin_edges = np.linspace(min_value, max_value, num_bins+1)  

    # Create a figure and axis object
    fig, ax = plt.subplots()

    # Plot both histograms with the same bin edges
    ax.hist(list1, bins=bin_edges, alpha=0.5, label=label1, edgecolor='black')
    ax.hist(list2, bins=bin_edges, alpha=0.5, label=label2, edgecolor='black')

    # Add labels, title, and legend
    ax.set_xlabel(xlabel)
    ax.set_ylabel('Count')
    ax.set_yscale('log')
    ax.set_title(title)
    ax.legend()

    # Save the plot
    fig.savefig(title+" Comparison.png")
    logging.info(f"Saved {title} Comparison plot to {title} Comparison.png")
    return None

def get_bbox_images(frame_list_orig, frame_list_processed, area_lb, area_ub, pad):
    """
    Processes a list of frames to extract bounding boxes and annotate images with bounding boxes.

    Args:
        frame_list_orig (List[np.ndarray]): List of original frames.
        frame_list_processed (List[np.ndarray]): List of processed frames.
        area_lb (int): Lower bound for contour area.
        area_ub (int): Upper bound for contour area.
        pad (int): Padding around bounding boxes.

    Returns:
        Tuple[List[np.ndarray], List[List[Tuple[int, int, int, int]]], List[List[np.ndarray]]]:
            - Annotated frames with bounding boxes.
            - List of bounding boxes for each frame.
            - List of contours for each frame.
    """
    frame_bb_list = []
    bbox_list = []
    fry_list = []

    for orig, processed in zip(frame_list_orig, frame_list_processed):
        # Extract padded bounding boxes and contours
        curr_bbox_list, curr_fry = get_padded_bbox_list(processed, area_lb, area_ub, pad)
        fry_list.append(curr_fry)

        # Annotate original image with bounding boxes
        annotated_image = draw_bounding_boxes(orig, curr_bbox_list, color=(200, 0, 255), thickness=2)
        if annotated_image.ndim == 2:  # Ensure the image has 3 channels
            annotated_image = np.repeat(annotated_image[:, :, np.newaxis], 3, axis=2)

        frame_bb_list.append(annotated_image)
        bbox_list.append(curr_bbox_list)

        # Save each bounding box as a separate image
        #import pdb; pdb.set_trace()
        #for bbox_idx, (x, y, w, h) in enumerate(curr_bbox_list):
        #    cropped_bbox = orig[max(0, y):y + h, max(0, x):x + w]  # Crop the bounding box region
        #    bbox_path = os.path.join(output_dir, f"frame_{frame_idx}_bbox_{bbox_idx}.png")
        #    cv2.imwrite(bbox_path, cropped_bbox)  # Save cropped bounding box as an image


    return frame_bb_list, bbox_list, fry_list



def get_padded_bbox_list(image_np, area_lb=24, area_ub=500, pad=10):
    """
    Extracts contours from an image and returns padded bounding boxes.

    Args:
        image_np (np.ndarray): Input image (grayscale or color).
        area_lb (int): Lower bound for contour area.
        area_ub (int): Upper bound for contour area.
        pad (int): Padding around bounding boxes.

    Returns:
        Tuple[List[Tuple[int, int, int, int]], List[np.ndarray]]:
            - List of padded bounding boxes (x, y, w, h).
            - List of contours corresponding to the bounding boxes.
    """
    if image_np.ndim == 3:
        image_np = cv2.cvtColor(image_np, cv2.COLOR_BGR2GRAY)

    # Extract contours
    contours = get_contours(image_np)
    filtered_contours = contour_filter_area(contours, operator.ge, area_lb)
    filtered_contours = contour_filter_area(filtered_contours, operator.le, area_ub)

    # Compute padded bounding boxes
    bbox_list = get_padded_bounding_boxes(filtered_contours, pad)
    return bbox_list, filtered_contours

    
def get_processed_images (frame_names, filter_list, src_dir):

    frame_orig_list = []
    frame_processed_list = []
    
    for curr_img_name in tqdm(frame_names):

        # Get current image
        curr_img_path = os.path.join(src_dir, curr_img_name)
        curr_img = cv2.imread(curr_img_path)

        # Apply series of filters to the image. 
        # Return list of tuples with filter name and filtered image
        # Filers are applied in sequence, 
        # so the nth image is the result of the first n filters
        processed_images = process_frame (curr_img, filter_list)
        orig_name, orig_image = processed_images[0]
        processed_name, processed_image = processed_images[-1]

        frame_orig_list.append(orig_image)
        frame_processed_list.append(processed_image)

    # Return list of original and fully processed images
    return (frame_orig_list, frame_processed_list)
        
def get_track_frames(frame_list_orig, frame_list_processed, bbox_list, dist_thresh=20):
    
    # List to store processed frames with tracking information
    track_images = []

    detected_objs = set()
    tracker = Tracker(distance_function="euclidean", distance_threshold=dist_thresh)

    for ctr, (curr_frame_orig, 
              curr_frame_processed, 
              curr_bboxes) in enumerate(zip(frame_list_orig, 
                                            frame_list_processed, 
                                            bbox_list)):

        # Get bbox centers
        bbox_centers = [np.array([x+w/2, y+h/2]) for (x,y,w,h) in curr_bboxes]

        # Convert detection bounding box centers to Norfair Detection objects
        detections = [Detection(points) for points in bbox_centers]
        
        # Update the tracker with the current detections
        tracked_objects = tracker.update(detections=detections)

        # Count tracked objects
        for curr in tracked_objects:
            detected_objs.add(curr.initializing_id)

        
        # Draw the tracked objects (with IDs and bounding boxes) on the frame
        curr_frame_track = np.copy(curr_frame_orig)
        draw_points(curr_frame_track, tracked_objects, color=(250,250,250))
        
        # Append the processed frame (with tracking info) to the results list
        track_image = np.copy(curr_frame_track)
        if track_image.ndim == 2:
            track_image = np.repeat(track_image[:, :, np.newaxis], 3, axis=2)
        track_images.append(np.copy(track_image))

    
    # Return the ordered list of processed images
    return track_images, detected_objs

def make_bbox_hist_plot(bbox_list, plot_title, xlabel,
                        num_bins=50 , xlim=None, ylog=False):

    #curr_out_dir = out_dir / Path("Movies") / Path(dataset) / Path(curr_movie_name)
    #curr_out_file = Path("_".join([chosen_filter, str(chosen_channel) + file_suffix]))
    fig, ax = plt.subplots()

    # Plot the data
    # Create the histogram on the axis 'ax'
    n, bins, patches = ax.hist(bbox_list, bins=num_bins, edgecolor='black')
    
    # Add labels and title
    ax.set_xlabel(xlabel)
    if xlim:
        ax.set_xlim(xlim)
    else:
        ax.set_xlim([0,500])
    ax.set_ylabel('Count')
    if ylog:
        ax.set_yscale('log')
    ax.set_title('Fry Bbox Histogram: ' + plot_title)
    
    # Add a legend
    ax.legend()
    plt.show()

def save_bbox_images(frame_list_bb, bbox_list, out_dir):
    """
    Save images with bounding boxes to disk.
    
    Parameters:
    frame_list_bb: List of images 
    bbox_list: List of bounding boxes
    out_dir: Directory to save images
    curr_movie_name: Name of the movie
    """

    bbox_dir = out_dir / Path("Bboxes")
    bbox_dir.mkdir(parents=True, exist_ok=True)
    for frame_ctr, (curr_frame, curr_bboxes) in enumerate(zip(frame_list_bb, bbox_list)):
        for bbox_ctr, (x, y, w, h) in enumerate(curr_bboxes):
            out_file = Path(f"bbox_{frame_ctr}_{bbox_ctr}_{x}_{y}.png")
            out_file_posix = bbox_dir / out_file
            cropped_bbox = curr_frame[max(0, y):y + h, 
                                      max(0, x):x + w]  # Crop the bounding box region
            cv2.imwrite(out_file_posix.as_posix(), cropped_bbox)  # Save cropped bounding box as an image
    
    return None

def make_movie_comparisons (frame_dict, filter_list, src_dir, 
                            movie_out_dir, analysis_params_dict, 
                            fps=3):
   
   # Initialize lists to store data for analysis
    num_fry_list = []
    bbox_areas_list = []
    bbox_aspect_ratios_list = []
    raw_fry_list = []


    # Process frames or each movie and save them to disk
    for curr_movie_name in frame_dict:

        # Create output file name and directory  
        curr_out_file = Path("_".join([filter_list[-1][1], curr_movie_name + '.mp4']))
        curr_out_dir = movie_out_dir / Path(curr_movie_name)
        curr_out_dir.mkdir(parents=True, exist_ok=True)

        frame_names = frame_dict[curr_movie_name]
        

        # Apply filters to the frames for the current movie
        # Return lists of original and processed images
        frame_list_orig, frame_list_processed = get_processed_images (frame_names, 
                                                                      filter_list, 
                                                                      src_dir)
        # Get processed images with bounding boxes, list of bounding boxes, 
        # and list of closed contours

        padding = analysis_params_dict["BoundingBoxParams"]["padding"]
        area_lb = analysis_params_dict["BoundingBoxParams"]["area_lb"]
        area_ub = analysis_params_dict["BoundingBoxParams"]["area_ub"]
        frame_list_bb, bbox_list, raw_fry_list = get_bbox_images(frame_list_orig, 
                                                                 frame_list_processed,
                                                                 area_lb=area_lb, 
                                                                 area_ub=area_ub, 
                                                                 pad=padding)

        save_bbox_images(frame_list_orig, bbox_list, curr_out_dir)
        logging.info(f"Saved bboxes for {curr_movie_name} to {curr_out_dir}")


        # Get processed images with tracking information and list of detected objects
        frame_list_track, detected_objs = get_track_frames(frame_list_bb, 
                                                           frame_list_processed, 
                                                           bbox_list, 
                                                           dist_thresh=20)

        num_fry_list = [len(x) for x in bbox_list]
        bbox_areas_list = [w*h for curr_bbox_list in bbox_list 
                           for (x,y,w,h) in curr_bbox_list]
        bbox_aspect_ratios_list = [w/(h+1e-9) for curr_bbox_list in bbox_list 
                                   for (x,y,w,h) in curr_bbox_list]


        for ctr, (orig, bb, track) in enumerate(zip(frame_list_orig, 
                                                    frame_list_bb, 
                                                    frame_list_track)):
            # Annotate the frames with text
            orig = rel_annotate_frame(orig, "Original", (.02,0.95) )
            orig = rel_annotate_frame(orig, "Image", (.02,0.975) )
            bb = rel_annotate_frame(bb, "Bounding Box", (.02,0.95) )
            track= rel_annotate_frame(track, "Bounding Box + ", (.02,0.95) )
            track= rel_annotate_frame(track, "Tracking", (.02,0.975) )

            # Write frame number on image
            curr_frame_num = ctr
            for frame in [orig, bb, track]:
                frame = rel_annotate_frame(frame, "Frame", (.865,0.925) )
                frame = rel_annotate_frame(frame, "Number", (.85,0.95) )
                frame = rel_annotate_frame(frame, str(curr_frame_num), (.90,0.975) )            

        # Create a video clip from the image frames
        clip_orig = ImageSequenceClip(frame_list_orig, fps=fps)  
        clip_bb = ImageSequenceClip(frame_list_bb, fps=fps)  
        clip_track = ImageSequenceClip(frame_list_track, fps=fps)  
    
        # Get the height of the clips (they should all have the same height)
        clip_height = clip_orig.h
    
        # Create a narrow vertical spacer (e.g., 10px wide) with the same height and black color
        spacer = ColorClip(size=(10, clip_height), color=(0, 0, 0), duration=clip_orig.duration)
    
        # Arrange the clips side by side with spacers in between
        final_clip = clips_array([[clip_orig, spacer, clip_bb, spacer, clip_track]])
    
        # Save the video clip
        out_file_posix = curr_out_dir / curr_out_file
        logging.info(f"Saving {out_file_posix}")
        logging.info(f"\n")

        # Plot the number of fry detected in each frame
        fry_summary(analysis_params_dict, num_fry_list, curr_out_dir)

        # Save the video clip
        final_clip.write_videofile(out_file_posix.as_posix())
        logging.info(f"Saved {curr_movie_name} movie to {out_file_posix}")
    return None
    
def process_frame (image_np, filter_list):
    """
        Processes an image by applying a series of filters.
    
        This function takes an image (as a NumPy array) and a list of filter functions
        along with their parameters. It applies each filter in sequence to the image,
        while storing the original and final processed images. 
    
        Args:
        - image_np: NumPy array representing the image to process.
        - filter_list: List of tuples where each tuple contains:
            (filter_function, filter_function_name, params)
            - filter_function: The function to apply to the image.
            - filter_function_name: A string representing the name of the filter.
            - params: A dictionary of parameters to pass to the filter function.
    
        Returns:
        - img_list: A list of tuples containing:
            (filter_name, processed_image)
            - The original image and the final processed image after all filters are applied.
    """
    # Ensure not info logging, if user wants to avoid too many messages
    logger = logging.getLogger()
    orig_log_level = logger.getEffectiveLevel()
    logger.setLevel(logging.WARN)
        
    img_list = [("Orig Image", image_np)]  # Store the original image
    for filter_func, filter_func_name, params in filter_list:
        image_np = filter_func(np.copy(image_np), **params)  # Apply the filter function
        
    img_list.append((filter_func_name, image_np))

    # Restore original logging level
    logger.setLevel(orig_log_level)
    # Return list with original & final images 
    # along with original & final filters (TBD: Remove filter info)
    return img_list


def plot_counts(list1, title, ylim, out_dir, fname):
    """
    Plot number of fry
    
    Parameters:
    list1: List of fry counts
    """
    
    # Create a figure and axis object
    fig, ax = plt.subplots()

    # Plot both histograms with the same bin edges
    ax.plot(list1)

    # Add labels, title, and legend
    ax.set_xlabel("Frame Number")
    ax.set_ylabel('Average Fry Count')
    ax.set_ylim(ylim)
    ax.set_title(title)

    # Save the plot
    fname_posix = out_dir / Path(fname)
    fig.savefig(fname_posix.as_posix())
    logging.info(f"Saved {title} plot to {fname}")
    return None

def fry_summary(analysis_params_dict, num_fry, out_dir):

    window_size = analysis_params_dict["RollingAverage"]["window_size"]
    min_max_clip = analysis_params_dict["RollingAverage"]["min_max_clip"]
    ra = centered_rolling_average(num_fry, window_size, min_max_clip)

    ymin = analysis_params_dict["PlotParams"]["ymin"]
    ymax = analysis_params_dict["PlotParams"]["ymax"]
    plot_counts(ra,"Fry (Rolling Avg)",  ylim=[ymin, ymax],
                out_dir=out_dir, fname="Fry_Rolling_Avg.png")
    return None

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Process a video file and save the results')
    parser.add_argument('-f', '--filter_config_file', type=str, 
                        default = 'filters.cfg', help='Path to the filters config file')
    parser.add_argument('-i', '--io_config_file', type=str, 
                        default = 'io_dirs.cfg', help='Path to the io config file')
    parser.add_argument('-a', '--analysis_config_file', type=str, 
                        default = 'analysis.cfg', help='Path to the analysis config file')
    args = parser.parse_args()

    # Set up logging
    logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")

    # Assign names of config files
    filter_config_file = args.filter_config_file
    io_config_file = args.io_config_file
    analysis_config_file = args.analysis_config_file

    # Get filter functions, input/output directories, and analysis parameters
    filter_list = get_filter_functions(filter_config_file)
    io_dir_dict = get_input_output_dirs(io_config_file)
    analysis_params_dict = get_analysis_params(analysis_config_file)
 
    # Get the frame dictionary - maps movie names to ordered frame names
    frame_dict = make_frame_dict(io_dir_dict["InputDirs"]["src_dir"])
    full_movie_list = list(frame_dict.keys()).sort()

    # Get the list of movies to process
    if io_dir_dict["InputDirs"]["movies"]:
        if isinstance(io_dir_dict["InputDirs"]["movies"][0], int):
            # Movies are specified by index
            test_dict = {full_movie_list[x]: frame_dict[full_movie_list[x]] 
                         for x in io_dir_dict["InputDirs"]["movies"]}
        elif isinstance(io_dir_dict["InputDirs"]["movies"][0], str):
            # Movies are specified by name
            test_dict = {x: frame_dict[x] for x in io_dir_dict["InputDirs"]["movies"]}
        else:
            raise ValueError("Invalid movie list in config file.")
    else:
        test_dict = frame_dict

    # Make movie comparing raw video, video with bounding boxes, and video with tracking
    make_movie_comparisons(test_dict, filter_list, io_dir_dict["InputDirs"]["src_dir"], 
                           io_dir_dict["OutputDirs"]["movie_output_root_dir"],
                           analysis_params_dict)
    
