import argparse
import cv2
import os
import torch

from collections import defaultdict
from moviepy.editor import *
from pathlib import Path
from tqdm import tqdm


# import some common detectron2 utilities
from detectron2.config import get_cfg
from detectron2.engine import DefaultPredictor
from detectron2.utils.visualizer import ColorMode, Visualizer

# Get ALDI imports
from aldi.config import add_aldi_config

class AttributeDict(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__dict__ = self

def get_predictor(expt_cfg, model_weights):
    '''Get Inference predictor for model with given config and weights'''
    cfg = get_cfg()
    cfg.set_new_allowed(True)
    add_aldi_config(cfg)
    cfg.merge_from_file(expt_cfg)
    cfg.MODEL.ROI_HEADS.SCORE_THRESH_TEST = 0.5  # set threshold for this model
    cfg.MODEL.WEIGHTS = model_weights
    predictor = DefaultPredictor(cfg)
    return predictor


def make_frame_dict(src_dir_str):
    '''Return dict mapping each movie to its ordered frames'''
    src_dir = Path(src_dir_str)

    frame_dict = defaultdict(list)
    frame_list = sorted([x.name for x in src_dir.iterdir() if 'ipynb' not in x.name ])
    
    # Group images by prefix label
    for curr_img in frame_list:
        prefix = "_".join(curr_img.split("_")[0:9])
        frame_dict[prefix].append(curr_img)
    
    # Sort images for each prefix number by what I believe to be frame number
    for curr_movie in sorted(frame_dict.keys()):
        try:
            frame_dict[curr_movie] = sorted(frame_dict[curr_movie], 
                                            key=lambda x:int(x.split("_")[-1][:-4]))
        except:
            print(curr_movie,"---")
    return frame_dict

def recover_pred_frame_dict(root_dir_str):
    pred_frame_dict = defaultdict(list)
    root_dir = Path(root_dir_str)
    movie_list = sorted([x for x in root_dir.iterdir() if 'ipynb' not in x.name ])
    for curr_movie in movie_list:
        try:
            curr_frames = [x.name for x in curr_movie.iterdir()] 
        except:
            print(curr_movie,"---")
            continue
        curr_path = root_dir / curr_movie
        curr_name = "/".join(curr_path.parts)[1:]
        temp = sorted(curr_frames, key=lambda x:int(x.split("_")[-1][:-4]))
        pred_frame_dict[curr_name] = [os.path.join(curr_name,x) for x in temp]
        
    return pred_frame_dict


def make_pred_frame_dict(frame_dict, predictor, src_dir, out_dir):
    '''Return dict mapping each movie to its ordered prediction frames'''
    pred_frame_dict = defaultdict(list)
    num_movies = len(frame_dict)
    bad_file_list = []
    thing_colors = [(175, 175, 255)]
    metadata = AttributeDict({"thing_classes": ["Fish"],  
                              "thing_colors": thing_colors })
    for ctr, curr_movie_name in enumerate(frame_dict):
        curr_out_dir = out_dir / Path("Pred_Frames") / Path(curr_movie_name)
        curr_out_dir.mkdir(parents=True, exist_ok=True)
        print(curr_out_dir, (ctr+1), " of ", num_movies)
        for curr_img_name in tqdm(frame_dict[curr_movie_name]):

            # Get predictions for current image
            curr_img_path = os.path.join(src_dir, curr_img_name)
            im = cv2.imread(curr_img_path)
            try:
                outputs = predictor(im)
            except:
                print("Error with ", curr_img_name)
                bad_file_list.append(curr_img_name)
                continue

            # Draw image
            v = Visualizer(im[:, :, ::-1], metadata=metadata, scale=1.2, instance_mode=ColorMode.SEGMENTATION)
            out = v.draw_instance_predictions(outputs["instances"].to("cpu"))

            # Get image dimensions
            out_img = out.get_image()[:, :, ::-1]
            _, width, _ = out_img.shape

            # Calculate center coordinates for top-center position
            text_center_x = int(width / 2)
            text_center_y = 50  # Adjust y-coordinate as needed (distance from top edge)

            # Text parameters
            count = torch.gt(outputs['instances'].scores, 0).sum().item()
            text = "Number of Fish: " + str(count)
            fontFace = cv2.FONT_HERSHEY_SIMPLEX
            fontScale = 1.0
            color = (255, 175, 175)  
            thickness = 2

            # Get text size using getTextSize
            (text_width, text_height), _ = cv2.getTextSize(text, fontFace, fontScale, thickness)

            # Adjust center_y based on text height for better alignment
            text_center_y = text_center_y + text_height // 2  # Add half of text height

            # Bottom-left corner coordinates relative to center
            text_org = (text_center_x - text_width // 2, text_center_y)  

            # Write text on image
            out_img=out_img.copy()
            cv2.putText(out_img, text, text_org, 
                        fontFace, fontScale, 
                        color, thickness, cv2.LINE_AA)

            # Save image            
            outfile = curr_out_dir / Path("Pred_" + curr_img_name)
            cv2.imwrite('/'.join(outfile.parts),out_img)
            pred_frame_dict[curr_movie_name].append('/'.join(outfile.parts))
    print("Bad Files: ", bad_file_list )
    return pred_frame_dict


def make_movies(pred_frame_dict, out_dir, dataset, fps=10):
    '''Create movies from predicted frames and save to out_dir'''

    # Get list of movies
    movie_list = [x for x in pred_frame_dict]
    num_movies = len(movie_list)

    # Make output directory
    out_dir_root = out_dir / Path('movies') / Path(dataset)
    out_dir_root.mkdir(parents=True, exist_ok=True)
    #import pdb; pdb.set_trace()

    for ctr,curr_movie in enumerate(movie_list):
        print(curr_movie,"   ", ctr+1 ," of ", num_movies)
        curr_movie_name = curr_movie.split("/")[-1]
        outfile_path = out_dir_root / Path(curr_movie_name + ".mp4")
        if outfile_path.exists():
            print(f" {outfile_path.name}  already exists")
            continue
        outfile = '/'.join(outfile_path.parts)
    
        # List image filenames
        images = pred_frame_dict[curr_movie]
        if len(images) == 0:
            print("     empty")
            break
    
        # Define clip duration (assuming all images have same duration)
        clip_duration = 1/fps  # Adjust clip duration per image (in seconds)
        
        # Create video clips from images
        clips = [ImageClip(image).set_duration(clip_duration) 
                 for image in images ]
    
        # Concatenate clips
        concat_clip = concatenate_videoclips(clips, method="compose")
        concat_clip.fps = fps

        print(f"Saving {outfile}")
        concat_clip.write_videofile(outfile,logger=None,codec = 'libx264')


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--dataset", type=str, default='cfc_val')
    parser.add_argument("--data_root_dir", type=str, default='./datasets/cfc_daod/images')
    parser.add_argument("--model_cfg", type=str, 
                        default='./configs/cfc/Base-RCNN-FPN-CFC_strongaug_ema.yaml')
    parser.add_argument("--model_weights", type=str, 
                        default= './output/cfc/cfc_baseline_strongaug_ema/cfc_val_model_best.pth')
    parser.add_argument("--output_root_dir", type=str, 
                        default='/media/smgutstein/local_repos/Fish_Data_2.0/CalTech/caltech-fish-datasets/Results/Movies_Temp')
    parser.add_argument("--recover_dir", type=str, 
                        default="/media/smgutstein/local_repos/Fish_Data_2.0/CalTech/caltech-fish-datasets/Results/Movies_Temp")
    parser.add_argument('-t', '--test', action='store_true')
    parser.add_argument('-r', '--recover_pred_frames', action='store_true')
    args = parser.parse_args()


    if not args.recover_pred_frames:
        # Get predictor
        predictor = get_predictor(args.model_cfg, args.model_weights)

        # Make output directory
        out_dir = Path(args.output_root_dir) / Path(args.dataset)
        out_dir.mkdir(parents=True, exist_ok=True)

        # Get frame dict
        src_dir = os.path.join(args.data_root_dir, args.dataset)
        frame_dict = make_frame_dict(src_dir)

        # Limit number of movies for testing
        if args.test:
            frame_dict = {k: frame_dict[k] for k in list(frame_dict)[0:5]}

        # Get predicted frame dict
        pred_frame_dict = make_pred_frame_dict(frame_dict, predictor, src_dir, out_dir)
    else:
        # Recover predicted frame dict
        print("Recovering predicted frames and resuming movie making...")
        recover_dir = os.path.join(args.recover_dir, args.dataset, 'Pred_Frames')
        pred_frame_dict = recover_pred_frame_dict(recover_dir)

    make_movies(pred_frame_dict, args.output_root_dir, args.dataset, fps=10)