import cv2
import logging
import numpy as np
import operator

from collections import defaultdict
from moviepy.editor import ColorClip, clips_array, ImageClip, ImageSequenceClip
from moviepy.video.compositing.concatenate import concatenate_videoclips
from pathlib import Path
from tqdm import tqdm

logging.basicConfig(level=logging.INFO)

def make_frame_dict(src_dir_str):
    '''Return dict mapping each movie to its ordered frames'''
    src_dir = Path(src_dir_str)

    frame_dict = defaultdict(list)
    frame_list = sorted([x.name for x in src_dir.iterdir() 
                         if 'ipynb' not in x.name and x.is_file()])
    
    # Group images by prefix label
    for curr_img in frame_list:
        prefix = "_".join(curr_img.split("_")[0:9])
        frame_dict[prefix].append(curr_img)
    
    # Sort images for each prefix number by what I believe to be frame number
    for curr_movie in sorted(frame_dict.keys()):
        try:
            frame_dict[curr_movie] = sorted(frame_dict[curr_movie], 
                                            key=lambda x:int(x.split("_")[-1][:-4]))
        except:
            print(curr_movie,"---")
    return frame_dict

def process_single_frame_bboxes(image_path, filter_list, show_filters=False):
    """
    Applies a series of filters to an image and optionally returns intermediate results.

    Parameters:
    image_np (np.ndarray): The input image as a 2D or 3D NumPy array.
    filter_list (list): List of filters to apply. Each filter is a tuple (filter_func, filter_func_name, params).
    show_filters (bool): Whether to return intermediate images after each filter.

    Returns:
    list  a list of (filter_name, filtered_image) tuples. If show_filters=True,
    it returns entire series of processed images. Otherwise it only returns the 
    original image and the fully processed result.
                       
    """

    if isinstance(image_path, Path) or isinstance(image_path, str):
        image_np = cv2.imread(image_path)
    else:
        raise ValueError("Expected a string or Path variable for image file.")

    # Ensure not info logging, if user wants to avoid too many messages
    logger = logging.getLogger()
    orig_log_level = logger.getEffectiveLevel()
    if not show_filters:
        logger.setLevel(logging.WARN)
        
    img_list = [("Orig Image", image_np)]  # Store the original image

    # Apply each filter in the list
    #print(image_np.dtype, np.min(image_np), np.max(image_np))
    for filter_func, filter_func_name, params in filter_list:
        image_np = filter_func(np.copy(image_np), **params)  # Apply the filter function
        if show_filters:
            img_list.append((filter_func_name, image_np))  # Store intermediate result
   
    # Ensure final processed image included in img_list
    if not show_filters:
        img_list.append((filter_func_name, image_np))

    # Restore original logging level
    logger.setLevel(orig_log_level)
    return img_list


def process_single_frame(image_path, filter_list, show_filters=False):
    """
    Applies a series of filters to an image and optionally returns intermediate results.

    Parameters:
    image_np (np.ndarray): The input image as a 2D or 3D NumPy array.
    filter_list (list): List of filters to apply. Each filter is a tuple (filter_func, filter_func_name, params).
    show_filters (bool): Whether to return intermediate images after each filter.

    Returns:
    list  a list of (filter_name, filtered_image) tuples. If show_filters=True,
    it returns entire series of processed images. Otherwise it only returns the 
    original image and the fully processed result.
                       
    """

    if isinstance(image_path, Path) or isinstance(image_path, str):
        image_np = cv2.imread(image_path)
    else:
        raise ValueError("Expected a string or Path variable for image file.")

    # Ensure not info logging, if user wants to avoid too many messages
    logger = logging.getLogger()
    orig_log_level = logger.getEffectiveLevel()
    if not show_filters:
        logger.setLevel(logging.WARN)
        
    img_list = [("Orig Image", image_np)]  # Store the original image

    # Apply each filter in the list
    #print(image_np.dtype, np.min(image_np), np.max(image_np))
    for filter_func, filter_func_name, params in filter_list:
        image_np = filter_func(np.copy(image_np), **params)  # Apply the filter function
        if show_filters:
            img_list.append((filter_func_name, image_np))  # Store intermediate result
            logging.info(f"{filter_func_name}, {image_np.dtype}, {np.min(image_np)}, {np.max(image_np)},{np.sum(image_np ==0)},{np.sum(image_np ==255)}")
    
    # Ensure final processed image included in img_list
    if not show_filters:
        img_list.append((filter_func_name, image_np))
        logging.info(f"{filter_func_name}, {image_np.dtype}, {np.min(image_np)}, {np.max(image_np)},{np.sum(image_np ==0)},{np.sum(image_np ==255)}")

    # Restore original logging level
    logger.setLevel(orig_log_level)
    return img_list


def rel_annotate_frame(image, text, relative_position=(.05,0.95), relative_font_size=0.001):
    """ Add text annotations to image using positions relative to image dimensions"""
    height, width = image.shape[:2]  # Get the image dimensions
    
    # Calculate the absolute font size and position relative to the image dimensions
    font_scale = relative_font_size * min(width, height)  # Scale font size based on image size
    x_pos = int(relative_position[0] * width)   # X position as a fraction of image width
    y_pos = int(relative_position[1] * height)  # Y position as a fraction of image height

    # Define other parameters for cv2.putText
    font = cv2.FONT_HERSHEY_SIMPLEX
    color = (255, 200, 200)  # White color (change as needed)
    thickness = 1
    
    # Add the text to the image
    cv2.putText(image, text, (x_pos, y_pos), font, font_scale, 
                color, thickness, cv2.LINE_AA)

    return image

def get_contours(image):

    # Find contours of the connected components
    contours, _ = cv2.findContours(image, 
                                   cv2.RETR_EXTERNAL, 
                                   cv2.CHAIN_APPROX_SIMPLE)

    return contours

def get_bounding_boxes(contours):

    bbox_list = []
    for i, contour in enumerate(contours):
        # Get the bounding box coordinates
        x, y, w, h = cv2.boundingRect(contour)
        
        bbox_list.append((x,y,w,h))
    return bbox_list

def get_bounding_box_areas(contours):

    bbox_list = []
    for i, contour in enumerate(contours):
        # Get the bounding box coordinates
        x, y, w, h = cv2.boundingRect(contour)
        
        #bbox_list.append(x*y)
        bbox_list.append(w*h)
    #bbox_list = sorted(bbox_list)
    return bbox_list


def get_bounding_box_aspect_ratios(contours):

    bbox_list = []
    for i, contour in enumerate(contours):
        # Get the bounding box coordinates
        x, y, w, h = cv2.boundingRect(contour)
        
        #bbox_list.append(x/(y+1e-6))
        bbox_list.append(w/(h+1e-6))
    bbox_list = sorted(bbox_list)
    return bbox_list

def get_padded_bounding_boxes(contours, pad=10):
    """
    Return padded bounding boxes for list of closed contours
    Parameters:
    contour (list): List of detected contours, where each contour is represented as an array of points
    pad (int): Amount of padding to add to bounding boxes
    
    Returns:
    bbox_list (list): List of padded bounding boxes as 4-tuples in form (x,y,w,h)
    """

    bbox_list = []
    for i, contour in enumerate(contours):
        # Get the bounding box coordinates
        x, y, w, h = cv2.boundingRect(contour)

        # Pad the bounding box coordinates
        x -= pad
        y -= pad
        w += 2*pad
        h += 2*pad
        bbox_list.append((x,y,w,h))
        
    return bbox_list

def draw_bounding_boxes(image, bbox_list, color=(0, 255, 0), thickness=2):
    """
    Draw bounding boxes on a deep copy of the image.
    
    Parameters:
    image (np.ndarray): The original image (unaltered).
    bbox_list (list): List of bounding box coordinates (x, y, w, h).
    color (tuple): Color of the rectangle (BGR format). Default is green.
    thickness (int): Thickness of the rectangle border. Default is 2 pixels.
    
    Returns:
    np.ndarray: A new image with the bounding boxes drawn.
    """
    
    # Create a deep copy of the image to avoid modifying the original
    image_copy = np.copy(image)
    
    # Loop through each bounding box in the list
    for bbox in bbox_list:
        x, y, w, h = bbox  # Unpack the bounding box coordinates
        # Draw the rectangle on the copied image
        cv2.rectangle(image_copy, (x, y), (x + w, y + h), color, thickness)
    
    return image_copy

    
def contour_filter_area(contours, comp_func, ref_area):
    filtered_contours = []
    for curr in contours:
        x, y, w, h = cv2.boundingRect(curr)
        #if x*y >= min_area:
        #if w*h >= min_area:
        if comp_func (w*h, ref_area):
            filtered_contours.append(curr)
    return filtered_contours

def contour_filter_aspect_ratio(contours, min_aspect_ratio):
    filtered_contours = []
    for curr in contours:
        x, y, w, h = cv2.boundingRect(curr)
        #if x/(y+1e-6) >= min_aspect_ratio:
        if w/(h+1e-6) >= min_aspect_ratio:
            filtered_contours.append(curr)
    return filtered_contours